import copy
from math import atan2, ceil, cos, floor, pi, sin, sqrt, tan

from .interval import *
from .myrandom import *
from .pyka3dmr import *
from .tools import *

# ACTIONS


def cango(player, x, y):
    z_player = player[P_Z]
    z_bottom, z_top = getBlockAltitudes(x, y)
    return z_player >= z_top - player_climbing_step(player) and not player[P_VZ] or z_player >= z_top or z_player + player_height(player) <= z_bottom


def get_avance_position(player, l, angle):
    l = bound_min_max(l, -1, 1) * player_step(player)
    x, y = player[:P_Z]
    player[PREV_X]=x
    player[PREV_Y]=y  
    player[NEXT_X]=x
    player[NEXT_Y]=y 
    dx, dy = cos_sin(angle)
    nx = x + l*dx
    ny = y + l*dy
    if (nx != x or ny != y) and cango(player, nx, ny):
        pr = player_radius(player)
        if cango(player, nx + pr*sign(dx), y):
            x = nx
            player[NEXT_X]=nx
        #else:
        #    player[NEXT_X]=x
        if cango(player, x, ny + pr*sign(dy)):
            y = ny
            player[NEXT_Y]=ny
        #else:
        #    player[NEXT_Y]=y
    #print(player)
    return x, y, dx, dy

# PLAYERS ACCESSORS


P_X, P_Y, P_Z, P_ANGLE, P_VZ, P_LEVEL, P_SCORE_IA, PREV_X, PREV_Y, NEXT_X, NEXT_Y = range(11)

DEFAULT_PLAYER_HEIGHT = 2
DEFAULT_PLAYER_RADIUS = .15
DEFAULT_PLAYER_FACTOR = 1 / 3
Z_XY_RATIO = 10

FusionList=[]
NbFusion=0


def player_height(player): return DEFAULT_PLAYER_HEIGHT + \
    (player[P_LEVEL] - 1)//2*DEFAULT_PLAYER_HEIGHT*DEFAULT_PLAYER_FACTOR
def player_radius(player): return DEFAULT_PLAYER_RADIUS + \
    (player[P_LEVEL] - 2)//2*DEFAULT_PLAYER_RADIUS*DEFAULT_PLAYER_FACTOR


def player_climbing_step(player): return player_height(player) * 3 / 7
def player_eye_height(player): return player_height(player) * 5 / 7
def player_eye_altitude(player): return player[P_Z] + player_eye_height(player)


def player_jump(player): return 3 * \
    (player_radius(player) + player_height(player)/14)


def player_step(player): return sqrt(
    dist2(player_radius(player) / 2, player_eye_height(player) / Z_XY_RATIO / 2))


def player_attract_sensor(player): return max(player_step(player), sqrt(
    dist2(player_radius(player), player_eye_height(player) / Z_XY_RATIO)))


def player_repulse_sensor(player): return max(player_attract_sensor(player), dist2(
    1 + player_radius(player)*2, 1 + player_height(player)/Z_XY_RATIO) - 1)


def levelup(player):
    block_altitude_bottom, block_altitude_top = getBlockAltitudes(
        player[P_X], player[P_Y])
    player[P_LEVEL] += 1
    if player[P_Z] < block_altitude_top and player[P_Z] + player_height(player) > block_altitude_top:
        player[P_LEVEL] -= 1


# MAP ACCESSORS
B_BOTTOM, B_HEIGHT, B_TOP, B_TEXTURE, B_COLOR = range(5)


def getBlockTop(x, y): return getBlockBottom(x, y) + getBlockHeight(x, y)


def getBlockAltitudes(x, y): return getBlockBottom(x, y), getBlockTop(x, y)


def getBlockInfos(x, y):
    a, h, it = getBlockBottom(x, y), getBlockHeight(
        x, y), getBlockTextureIndex(x, y)
    return a, h, a + h, it, (0, 0, 0)


class Pyka3D:
    def __init__(self):
        self.random = MyRandom()
        self.ok = 0
        self.id_action = 0
        self.init()
        self.action_fcts = (self.action_forward, self.action_strife_right, self.action_backward,
                            self.action_strife_left, self.action_pause, self.action_turn_left, self.action_turn_right, self.action_jump)

    def init(self):
        global FusionList, NbFusion
        FusionList=[]
        NbFusion=0
        
        self.random.seed(20100)  # Hello Vincent
        self.map_cp = [(2, 2), (2, 26), (26, 2), (26, 26), (14, 14)]
        self.liste_actions_clavier = []
        
        nplayers = 100
        self.main_player = [39.35, 30.3, 47, -3*pi/4, 0, 1, 1000, -1, -1, -1, -1]
        for k in range(2):
            levelup(self.main_player)
        self.players = [self.main_player]
        for k in range(nplayers - 1):
            # populates 3 zones : Tri-Maze (1/2) + Pikastle (1/3) + Heroes Arena
            x, y = k <= nplayers / 2 and (self.random.uniform(42, MAP_WIDTH - 3), self.random.uniform(5, MAP_HEIGHT - 4)) or k < 5 * nplayers / 6 and (
                self.random.uniform(2, 27), self.random.uniform(2, MAP_HEIGHT - 2)) or (self.random.uniform(28, 41), self.random.uniform(4, MAP_HEIGHT - 4))
            block_altitude_bottom, block_altitude_top = getBlockAltitudes(x, y)
            player = [x, y, block_altitude_bottom >= DEFAULT_PLAYER_HEIGHT and self.random.random() < 2/3 and self.random.uniform(
                0, block_altitude_bottom - DEFAULT_PLAYER_HEIGHT) or self.random.uniform(block_altitude_top, 255), self.random.uniform(0, 2*pi), 0, 1, 0, 0, 0, 0, 0]
            for k in range(int(self.random.uniform(0, 4))):
                levelup(player)
            self.players.append(player)
        self.ok = 1

    def get_cp(self):
        return self.map_cp

    def get_players(self):
        return self.players
    
    def get_Fusion(self):
        return FusionList
    
    def clear_Fusion(self, number):
        global FusionList, NbFusion
        
        print("CLEAR : Before : ", NbFusion, FusionList)
        FusionListNew = []
        for i in range(len(FusionList)-number):
            FusionListNew.append(FusionList[i])
        print("CLEAR : Now : ", FusionListNew)              

        FusionList = []        
        FusionList.extend(FusionListNew)
        NbFusion = len(FusionListNew)
        print("CLEAR : After : ", NbFusion, FusionList)


    def ia(self ):       
        global FusionList, NbFusion
        
        for player in self.players:
            main = player == self.main_player

            # get nearest player
            p2near, d2near = None, dist2(MAP_WIDTH, MAP_HEIGHT)
            for p2id in range(len(self.players)):
                p2 = self.players[p2id]
                if p2 != player:
                    td2 = dist2(p2[P_X], p2[P_Y], player[P_X], player[P_Y])
                    if td2 < d2near:
                        d2near = td2
                        p2near = p2
                        p2nearid = p2id

            # handle gravity
            x, y, z = player[:P_ANGLE]
            block_altitude_bottom, block_altitude_top = getBlockAltitudes(
                x, y)
            minz = z >= block_altitude_top and block_altitude_top
            ph = player_height(player)
            if player[P_VZ]:
                maxz = z + \
                    ph >= block_altitude_top and 255 or max(
                        0, block_altitude_bottom - ph)
                z = bound_min_max(z + player[P_VZ], minz, maxz)
                player[P_Z] = z
                if z == minz or z == maxz:
                    player[P_VZ] = 0
            if z != minz:
                player[P_VZ] -= 1

            if p2near:  # handle interactions with nearest player
                if d2near <= player_radius(player)**2 and player[P_Z] <= p2near[P_Z] + player_height(p2near) and player[P_Z] + ph >= p2near[P_Z]:
                    if player[P_LEVEL] >= p2near[P_LEVEL] and p2near != self.main_player:  # fusion
                    
                        print("Fusion level "+str(player[P_LEVEL])+" et " + str(p2near[P_LEVEL]))
                        FusionList.append([player[P_LEVEL],p2near[P_LEVEL],player[P_X],player[P_Y]])
                        
                        self.players.remove(p2near)
                        if main:
                            player[P_SCORE_IA] += 100 / \
                                (1 + self.main_player[P_LEVEL] -
                                    p2near[P_LEVEL])
                        levelup(player)
                    elif player[P_LEVEL] > p2near[P_LEVEL] and p2near == self.main_player:  # attack
                        p2near[P_SCORE_IA] -= player[P_LEVEL] - \
                            p2near[P_LEVEL]
                        player[P_SCORE_IA] = 0

            # handle interactions with special blocks
            # evolution-stone
            if main and not player[P_VZ] and z == block_altitude_top:
                x, y = floor(x), floor(y)
                try:
                    i = self.map_cp.index((x, y))
                    for player in self.players[1:]:
                        levelup(player)
                    self.map_cp.pop(i)
                except:
                    pass

            if not main:  # random + reactive ia actions
                mode = player[P_SCORE_IA]
                player[P_SCORE_IA] = (p2near[P_Z] >= player[P_Z] and p2near[P_Z] <= player[P_Z] + player_height(player) or player[P_Z] >= p2near[P_Z] and player[P_Z] <= p2near[P_Z] + player_height(p2near)) and (d2near <= player_repulse_sensor(player)**2 and (
                    player[P_LEVEL] < p2near[P_LEVEL] or p2near == self.main_player and player[P_LEVEL] == p2near[P_LEVEL]) and -1 or d2near <= player_attract_sensor(player)**2 and (player[P_LEVEL] > p2near[P_LEVEL] or p2near != self.main_player and player[P_LEVEL] == p2near[P_LEVEL]))

                s = not mode or d2near >= player_step(
                    player)**2 and 1 or d2near / player_step(player)**2

                nxs = []
                nys = []
                nds = []
                i = 0

                for k in range(mode and 4 or 1):
                    nx, ny, dx, dy = get_avance_position(
                        player, s, player[P_ANGLE] + k*pi/2)
                    nds.append(dist2(p2near[P_X], p2near[P_Y], nx, ny))
                    nxs.append(nx)
                    nys.append(ny)

                if mode:
                    i_max, i_min = 0, 0
                    nd_max, nd_min = nds[0], nds[0]
                    for k in range(1, len(nds)):
                        v = nds[k]
                        if v < nd_min:
                            i_min = k
                            nd_min = v
                        elif v > nd_max:
                            i_max = k
                            nd_max = v

                    i = i_min
                    if mode < 0:
                        i = i_max

                if nxs[i] >= 0 and nxs[i] < MAP_WIDTH and nys[i] >= 0 and nys[i] < MAP_HEIGHT and ((mode > 0 and nds[i] < d2near or mode < 0 and nds[i] > d2near) and abs(dprinc(player[P_ANGLE], p2near[P_ANGLE], pi/2)) < pi/8 or not mode and self.random.random() < .9):
                    self.action_fcts[i](player, s)
                elif mode and player[P_ANGLE] != p2near[P_ANGLE] or self.random.random() < .5:
                    self.action_turn_left(
                        player, mode and p2near[P_ANGLE] - player[P_ANGLE] or self.random.uniform(0, 2*pi))
                else:
                    self.action_jump(player, 1)

    def play(self, fid, params):
        self.ok = fid < len(self.action_fcts)
        param = 0
        if self.ok:
            param = self.action_fcts[fid](self.main_player, params)
        return param

    def save(self):
        return copy.deepcopy((self.liste_actions_clavier, self.map_cp, self.players, self.main_player, self.id_action, self.ok, self.random.rand_seed))

    def load(self, save):
        self.liste_actions_clavier, self.map_cp, self.players, self.main_player, self.id_action, self.ok, self.random.rand_seed = save

    def copy(self):
        p = Pyka3D()
        p.load(self.save())
        return p

    def action_copy(self, l):
        p = self.copy()
        p.action(l)
        return p

    def action(self, l):
        steps = 0
        while self.ok and (len(l)):
            t = len(l)

            # next action
            fid, params = t and l[0] or not t and None, t and l[1]

            self.ia( )
            param = self.play(fid, params)

            if t:
                if param:
                    l[1] = param
                else:
                    l.pop(1)
                    l.pop(0)
            steps += 1
        return steps

    def get_main_player(self):
        return self.main_player

    def get_liste_actions_clavier(self):
        return self.liste_actions_clavier

    def randomanip(self, v):
        while self.random.uniform(0, v) < .5 * v:
            pass

    def action_forward(self, player, params):
        param = bound_min_max(params, -1, 1)
        self.randomanip(abs(param) % 1)
        x0, y0 = player[:P_Z]
        x, y, dx, dy = get_avance_position(player, param, player[P_ANGLE])
        if player == self.main_player:
            player[P_SCORE_IA] -= .01*abs(param) + .001
        if x != x0 or y != y0:
            player[:P_Z] = (x, y)
            z_player = player[P_Z]
            z_top = getBlockTop(x, y)
            if z_player < z_top and z_player >= z_top - player_climbing_step(player):
                player[P_Z] = z_top
        return params - param

    def action_backward(self, player, l):
        self.random.random()
        return -self.action_forward(player, -l)

    def action_turn_right(self, player, param):
        for k in range(2):
            self.random.random()
        if player == self.main_player:
            player[P_SCORE_IA] -= .0001
        self.randomanip((dprinc(abs(param)) + pi) / 2 / pi)
        player[P_ANGLE] = dprinc(player[P_ANGLE] + param)
        return 0

    def action_turn_left(self, player, param):
        self.random.random()
        return self.action_turn_right(player, -param)

    def action_strife_left(self, player, params):
        self.random.random()
        self.action_turn_left(player, pi / 2)
        param = self.action_forward(player, params)
        self.action_turn_right(player, pi / 2)
        if player == self.main_player:
            player[P_SCORE_IA] += .0002
        self.random.random()
        return param

    def action_strife_right(self, player, params):
        for k in range(2):
            self.random.random()
        self.action_turn_right(player, pi / 2)
        param = self.action_forward(player, params)
        self.action_turn_left(player, pi / 2)
        if player == self.main_player:
            player[P_SCORE_IA] += .0002
        return param

    def action_jump(self, player, params):
        for k in range(6):
            self.random.random()
        params = abs(params)
        param = bound_min_max(params, 0, 1)
        if player == self.main_player:
            player[P_SCORE_IA] -= .0001 + param
        self.randomanip(param)
        if not player[P_VZ]:
            x, y, z_player = player[:P_ANGLE]
            z_top = getBlockTop(x, y)
            if z_player == z_top or not z_player:
                player[P_VZ] = player_jump(player)*param
        return params - param

    def action_pause(self, player, params):
        for k in range(7):
            self.random.random()
        if player == self.main_player:
            player[P_SCORE_IA] -= .0001
        return max(0, abs(params) - 1)


def jouer_selon(l, display):
    instance = Pyka3D()
    instance.init()
    steps = instance.action(l)

    return instance.get_main_player()[P_SCORE_IA], instance.get_liste_actions_clavier(), steps
