import json
import threading
from math import pi, cos, sin, tan, sqrt, ceil, atan2, floor
import tkinter as tk
from tkinter import *
from tkinter import ttk
from turtle import onkey
from datetime import datetime

from core.pyka3dlbcli import (P_LEVEL, P_X, P_Y, P_Z, P_ANGLE, P_VZ, P_SCORE_IA, PREV_X, PREV_Y, NEXT_X, NEXT_Y, getBlockAltitudes,
                              player_radius, player_step, player_jump, cango)
from core.pyka3dmr import MAP_HEIGHT, MAP_WIDTH

colorMouseHigh = "#FF00FF"
colorMouseLvl = "#00FFFF"
colorMouselvlminus1 = "#80FF00"
colorMouselvlminus2 = "#FFFF00"
colorMouselvlminus3 = "#FFAC00"
colorMouselvlminus4 = "#FF5400"
colorMouseLow = "#FFFFFF"
colorPlayer = "#0000FF"

reload = False
line = [(0,0,0)]
angle=0
move_length=0

ZOOM_FACTOR_MINIMAP = 18
EXTRA_ZOOM_ENTITIES = 5

ZOOM_FACTOR_ZOOMMAP = 35

ZOOM_FACTOR_ULTRAMAP = 120

NbFusion = 0

class App(Tk):
    def __init__(self, instance):
        global line
        self.pyka3d = instance
        Tk.__init__(self)
        self.actions = []
        self.history = []
        line = [(self.pyka3d.get_players()[0][0], self.pyka3d.get_players()[0][1] / 10, self.pyka3d.get_players()[0][2])]
        self.score = StringVar()
        self.level = StringVar()
        self.souris = StringVar()
        self.load_game = StringVar()
        self.angle = StringVar()
        self.dummy = ttk.Button(self, text="Dumb dumb dutton")
        self.angle.trace("w", self.ctrl_new_orientation_changed)
        self.move_length = StringVar()
        self.move_length.trace("w", self.ctrl_move_length_changed)
        self.nop_count = StringVar()
        self.keybinds = {
            'Left': self.ctrl_left,
            'Up': self.ctrl_forward,
            'Right': self.ctrl_right,
            '4': self.ctrl_move_left,
            'apostrophe': self.ctrl_move_left,
            'Down': self.ctrl_backward,
            '6': self.ctrl_move_right,
            'minus': self.ctrl_move_right,
            'Return': self.ctrl_jump,
            '5': self.ctrl_nop,            
            'parenleft': self.ctrl_nop,
            'u': self.ctrl_undo,
            'U': self.ctrl_undo
        }
        
        self.minimap = Canvas(self, width=(MAP_WIDTH+1) * ZOOM_FACTOR_MINIMAP, height=MAP_HEIGHT * ZOOM_FACTOR_MINIMAP+40, bg="black")
        self.zoommap = Canvas(self, width=12 * ZOOM_FACTOR_ZOOMMAP, height=12 * ZOOM_FACTOR_ZOOMMAP, bg="black")
        self.ultramap = Canvas(self, width=4 * ZOOM_FACTOR_ULTRAMAP, height=4 * ZOOM_FACTOR_ULTRAMAP, bg="black")
        self.output = tk.Text(self, height=10)
        self.outputfusions = tk.Text(self, height=10)

        self.minimap_pykas = []
        self.disable_updates = False
        self.init()
        self.init_minimap()
        self.update_minimap()
    
    def init(self):
        self.grid()
        row = 0
        ttk.Label(self, textvariable=self.score).grid(column=0, row=row, columnspan=1, sticky=EW)
        ttk.Label(self, textvariable=self.level).grid(column=1, row=row, columnspan=1, sticky=EW)
        ttk.Label(self, textvariable=self.souris).grid(column=2, row=row, columnspan=1, sticky=EW)

        row += 1
        ttk.Button(self, text="Wait", command=self.ctrl_nop).grid(column=0, row=row)
        ttk.Button(self, text="^", command=self.ctrl_forward).grid(column=1, row=row)
        ttk.Button(self, text="Jump", command=self.ctrl_jump).grid(column=2, row=row)

        row += 1
        ttk.Button(self, text="<", command=self.ctrl_left).grid(column=0, row=row)
        ttk.Button(self, text="v", command=self.ctrl_backward).grid(column=1, row=row)
        ttk.Button(self, text=">", command=self.ctrl_right).grid(column=2, row=row)

        row += 1
        ttk.Button(self, text="Undo", command=self.ctrl_undo).grid(column=0, row=row, columnspan=1, sticky=EW)
        ttk.Button(self, text="Dump list", command=self.ctrl_dump_list).grid(column=1, row=row, columnspan=1, sticky=EW)
        ttk.Button(self, text="Reset", command=self.ctrl_reset).grid(column=2, row=row, columnspan=1, sticky=EW)
        row += 1
        ttk.Label(self, text="Manual rotation").grid(column=0, row=row, columnspan=1, sticky=EW)
        ttk.Entry(self, textvariable=self.angle).grid(column=1, row=row, columnspan=1, sticky=EW)
        ttk.Button(self, text="Apply", command=self.ctrl_custom_rotate).grid(column=2, row=row)
        row += 1
        ttk.Label(self, text="Manual move").grid(column=0, row=row, columnspan=1, sticky=EW)
        ttk.Entry(self, textvariable=self.move_length).grid(column=1, row=row, columnspan=1, sticky=EW)
        ttk.Button(self, text="Apply", command=self.ctrl_custom_move).grid(column=2, row=row)
        row += 1
        ttk.Label(self, text="Manual nop").grid(column=0, row=row, columnspan=1, sticky=EW)
        ttk.Entry(self, textvariable=self.nop_count).grid(column=1, row=row, columnspan=1, sticky=EW)
        ttk.Button(self, text="Apply", command=self.ctrl_custom_nop).grid(column=2, row=row)
        row += 1
        ttk.Label(self, text="Load game").grid(column=0, row=row, columnspan=1, sticky=EW)
        ttk.Entry(self, textvariable=self.load_game).grid(column=1, row=row, columnspan=1, sticky=EW)
        ttk.Button(self, text="Load", command=self.ctrl_load_game).grid(column=2, row=row)

        row += 1
        HEIGHTENTRY = 3 
        self.output.grid(column=0, row=row, columnspan=3,  sticky=EW )

        row+=HEIGHTENTRY
        self.minimap.grid(column=0, row=row, columnspan=3, sticky=EW)


        
        
        self.bind("<Key>", self.ctrl_keyboard)

        self.zoommap.grid(column=3, row=0, rowspan=row, padx=5 )
        self.ultramap.grid(column=3, row=row, rowspan=1, padx=5 )

        self.outputfusions.grid(column=4, row=0, rowspan=row+1, sticky=NS)        

        self.ctrl_reset()
       
        l=[

            
        ]
       
        while len(l) != 0:
            a = l.pop(0)
            if isinstance(a, list) or isinstance(a, tuple):
                self.ctrl_action(a[0], a[1])
            else:
                self.ctrl_action(a, l.pop(0))
    
    
    def draw_pyka(self, pyka):
        plevel = self.pyka3d.get_main_player()[P_LEVEL]
        #r = player_radius(pyka) ** 2*EXTRA_ZOOM_ENTITIES
        if plevel == pyka[P_LEVEL]:
            colorscheme = colorMouseLvl
        elif plevel == pyka[P_LEVEL] + 1:
            colorscheme = colorMouselvlminus1
        elif plevel == pyka[P_LEVEL] + 2:
            colorscheme = colorMouselvlminus2      
        elif plevel == pyka[P_LEVEL] + 3:
            colorscheme = colorMouselvlminus3        
        elif plevel == pyka[P_LEVEL] + 4:
            colorscheme = colorMouselvlminus4
        elif plevel > pyka[P_LEVEL] + 4:
            colorscheme = colorMouseLow
        else:
            colorscheme = colorMouseHigh    
        
        self.minimap_pykas.append(self.minimap.create_rectangle((pyka[P_X] - 0.25) * ZOOM_FACTOR_MINIMAP, (pyka[P_Y] - 0.25) * ZOOM_FACTOR_MINIMAP, (pyka[P_X] + 0.25) * ZOOM_FACTOR_MINIMAP, (pyka[P_Y] + 0.25) * ZOOM_FACTOR_MINIMAP, fill=colorscheme, outline=""))
        
        #self.minimap_pykas.append(self.minimap.create_line(pyka[P_X] * ZOOM_FACTOR_MINIMAP, pyka[P_Y] * ZOOM_FACTOR_MINIMAP, (pyka[P_X]+2*cos(pyka[P_ANGLE])*player_step(pyka)*EXTRA_ZOOM_ENTITIES) * ZOOM_FACTOR_MINIMAP, (pyka[P_Y]+2*sin(pyka[P_ANGLE])*player_step(pyka)*EXTRA_ZOOM_ENTITIES) * ZOOM_FACTOR_MINIMAP, fill=colorscheme, width=2 ))
        self.minimap_pykas.append(self.minimap.create_line(pyka[P_X] * ZOOM_FACTOR_MINIMAP,
                                                           pyka[P_Y] * ZOOM_FACTOR_MINIMAP,
                                                           (pyka[P_X]+(pyka[NEXT_X]-pyka[PREV_X])*10) * ZOOM_FACTOR_MINIMAP,
                                                            (pyka[P_Y]+(pyka[NEXT_Y]-pyka[PREV_Y])*10) * ZOOM_FACTOR_MINIMAP, fill=colorscheme, width=2 ))
        
        self.minimap_pykas.append(self.minimap.create_text((pyka[P_X]+0.45) * ZOOM_FACTOR_MINIMAP, (pyka[P_Y]+0.45) * ZOOM_FACTOR_MINIMAP, fill=colorscheme, text=str(pyka[P_LEVEL]) ))
        
        if pyka[P_Z]<self.pyka3d.get_main_player()[P_Z]-1:
            self.minimap_pykas.append(self.minimap.create_line((pyka[P_X] - 0.25) * ZOOM_FACTOR_MINIMAP, (pyka[P_Y]) * ZOOM_FACTOR_MINIMAP, (pyka[P_X] + 0.25) * ZOOM_FACTOR_MINIMAP, (pyka[P_Y]) * ZOOM_FACTOR_MINIMAP, fill="black", width=4))
        elif pyka[P_Z]>=self.pyka3d.get_main_player()[P_Z]+1:
            self.minimap_pykas.append(self.minimap.create_line((pyka[P_X] - 0.25) * ZOOM_FACTOR_MINIMAP, (pyka[P_Y]) * ZOOM_FACTOR_MINIMAP, (pyka[P_X] + 0.25) * ZOOM_FACTOR_MINIMAP, (pyka[P_Y]) * ZOOM_FACTOR_MINIMAP, fill="black", width=4))
            self.minimap_pykas.append(self.minimap.create_line((pyka[P_X]) * ZOOM_FACTOR_MINIMAP, (pyka[P_Y] - 0.25) * ZOOM_FACTOR_MINIMAP, (pyka[P_X]) * ZOOM_FACTOR_MINIMAP, (pyka[P_Y] + 0.25) * ZOOM_FACTOR_MINIMAP, fill="black", width=4))
        #else:
        #    self.minimap_pykas.append(self.minimap.create_rectangle((pyka[P_X] - 0.35) * ZOOM_FACTOR_MINIMAP, (pyka[P_Y] - 0.35) * ZOOM_FACTOR_MINIMAP, (pyka[P_X] + 0.35) * ZOOM_FACTOR_MINIMAP, (pyka[P_Y] + 0.35) * ZOOM_FACTOR_MINIMAP, outline=colorscheme))
            


    def init_minimap(self):
        self.minimap.delete("all")
        for y in range(MAP_HEIGHT):
            for x in range(MAP_WIDTH):
                #bottom, top = getBlockAltitudes(x, y)
                if not cango(self.pyka3d.get_players()[0],x,y): #bottom < 0.01 < top:
                    self.minimap.create_rectangle(x * ZOOM_FACTOR_MINIMAP, y * ZOOM_FACTOR_MINIMAP, x * ZOOM_FACTOR_MINIMAP + ZOOM_FACTOR_MINIMAP, y * ZOOM_FACTOR_MINIMAP + ZOOM_FACTOR_MINIMAP, fill="grey", outline="")
                else:
                    self.minimap.create_rectangle(x * ZOOM_FACTOR_MINIMAP, y * ZOOM_FACTOR_MINIMAP, x * ZOOM_FACTOR_MINIMAP + ZOOM_FACTOR_MINIMAP, y * ZOOM_FACTOR_MINIMAP + ZOOM_FACTOR_MINIMAP, fill="black", outline="")


    def init_zoommap(self):
        self.zoommap.delete("all")
        
        main_player=self.pyka3d.get_players()[0];
        Xcenter = main_player[P_X]
        Ycenter = main_player[P_Y]
        X_CENTER_ZOOM_MAP = 6*ZOOM_FACTOR_ZOOMMAP
        Y_CENTER_ZOOM_MAP = 6*ZOOM_FACTOR_ZOOMMAP
        VECTOR_ZOOM = 200
        
        RadiusPlayer = max(int(ZOOM_FACTOR_ZOOMMAP*player_radius(main_player)),1)
  
        for u in range( int(Xcenter-7), int(Xcenter+7)+1 ):
            for v in range( int(Ycenter-7), int(Ycenter+7)+1 ):
                if ((u>=0 and u<65) and (v>=0 and v<31)):
                    if not cango(main_player,u,v): 
                        self.zoommap.create_rectangle(X_CENTER_ZOOM_MAP+(u-Xcenter)*ZOOM_FACTOR_ZOOMMAP,
                                                      Y_CENTER_ZOOM_MAP+(v-Ycenter)*ZOOM_FACTOR_ZOOMMAP,
                                                      X_CENTER_ZOOM_MAP+(u-Xcenter+1)*ZOOM_FACTOR_ZOOMMAP,
                                                      Y_CENTER_ZOOM_MAP+(v-Ycenter+1)*ZOOM_FACTOR_ZOOMMAP, fill="grey", outline="black")

        for player in self.pyka3d.get_players():
            if player==main_player:
                self.zoommap.create_oval( X_CENTER_ZOOM_MAP-RadiusPlayer, Y_CENTER_ZOOM_MAP-RadiusPlayer,
                                         X_CENTER_ZOOM_MAP+RadiusPlayer, Y_CENTER_ZOOM_MAP+RadiusPlayer, outline="#0000FF", width=2 )          
                
                self.zoommap.create_line( X_CENTER_ZOOM_MAP,
                                         Y_CENTER_ZOOM_MAP,
                                         X_CENTER_ZOOM_MAP+VECTOR_ZOOM*player_step(main_player)*cos(main_player[P_ANGLE]),
                                         Y_CENTER_ZOOM_MAP+VECTOR_ZOOM*player_step(main_player)*sin(main_player[P_ANGLE]), fill="#0000FF", width=2 ) 
                
                self.zoommap.create_text(X_CENTER_ZOOM_MAP+20, Y_CENTER_ZOOM_MAP+20, fill="#0000FF", text=str(main_player[P_LEVEL]))
                
                
            elif player!=main_player and (player[P_Z]==main_player[P_Z] or (player[P_Z]<=main_player[P_Z]+2 and player[P_VZ]>0)):
                deltaX0 = player[P_X]-Xcenter
                deltaY0 = player[P_Y]-Ycenter

                RadiusMice = int(ZOOM_FACTOR_ZOOMMAP*player_radius(player))+1
                if abs(deltaX0)<=8 and abs(deltaY0)<=8:
                    if main_player[P_LEVEL] == player[P_LEVEL]:
                        colorscheme = colorMouseLvl
                    elif main_player[P_LEVEL] == player[P_LEVEL] + 1:
                        colorscheme = colorMouselvlminus1
                    elif main_player[P_LEVEL] == player[P_LEVEL] + 2:
                        colorscheme = colorMouselvlminus2      
                    elif main_player[P_LEVEL] == player[P_LEVEL] + 3:
                        colorscheme = colorMouselvlminus3        
                    elif main_player[P_LEVEL] == player[P_LEVEL] + 4:
                        colorscheme = colorMouselvlminus4
                    elif main_player[P_LEVEL] > player[P_LEVEL] + 4:
                        colorscheme = colorMouseLow
                    else:
                        colorscheme = colorMouseHigh  
                        
                    self.zoommap.create_oval( X_CENTER_ZOOM_MAP+deltaX0*ZOOM_FACTOR_ZOOMMAP-RadiusMice,
                                             Y_CENTER_ZOOM_MAP+deltaY0*ZOOM_FACTOR_ZOOMMAP-RadiusMice,
                                                 X_CENTER_ZOOM_MAP+deltaX0*ZOOM_FACTOR_ZOOMMAP+RadiusMice,
                                                 Y_CENTER_ZOOM_MAP+deltaY0*ZOOM_FACTOR_ZOOMMAP+RadiusMice, outline=colorscheme, width=2 )     
                    
                    #self.zoommap.create_line( X_CENTER_ZOOM_MAP+deltaX0*ZOOM_FACTOR_ZOOMMAP,
                    #                         Y_CENTER_ZOOM_MAP+deltaY0*ZOOM_FACTOR_ZOOMMAP,
                    #                             X_CENTER_ZOOM_MAP+deltaX0*ZOOM_FACTOR_ZOOMMAP+VECTOR_ZOOM*player_step(player)*cos(player[P_ANGLE]),
                    #                             Y_CENTER_ZOOM_MAP+deltaY0*ZOOM_FACTOR_ZOOMMAP+VECTOR_ZOOM*player_step(player)*sin(player[P_ANGLE]), fill=colorscheme, width=2 ) 

                    self.zoommap.create_line( X_CENTER_ZOOM_MAP+deltaX0*ZOOM_FACTOR_ZOOMMAP,
                                             Y_CENTER_ZOOM_MAP+deltaY0*ZOOM_FACTOR_ZOOMMAP,
                                                 X_CENTER_ZOOM_MAP+deltaX0*ZOOM_FACTOR_ZOOMMAP+VECTOR_ZOOM*(player[NEXT_X]-player[PREV_X]),
                                                 Y_CENTER_ZOOM_MAP+deltaY0*ZOOM_FACTOR_ZOOMMAP+VECTOR_ZOOM*(player[NEXT_Y]-player[PREV_Y]), fill=colorscheme, width=2 ) 

                    
                    self.zoommap.create_text(X_CENTER_ZOOM_MAP+deltaX0*ZOOM_FACTOR_ZOOMMAP+20,
                                             Y_CENTER_ZOOM_MAP+deltaY0*ZOOM_FACTOR_ZOOMMAP+20,
                                             fill=colorscheme, text=str(player[P_LEVEL]))
 


    def init_ultramap(self):
        self.ultramap.delete("all")
       
        main_player=self.pyka3d.get_players()[0];
        Xcenter = main_player[P_X]
        Ycenter = main_player[P_Y]
        X_CENTER_ULTRA_MAP = 2*ZOOM_FACTOR_ULTRAMAP
        Y_CENTER_ULTRA_MAP = 2*ZOOM_FACTOR_ULTRAMAP
        VECTOR_ULTRA = 400
        
        
        RadiusPlayer = max(int(ZOOM_FACTOR_ULTRAMAP*player_radius(main_player)),1)
        RadiusPlayerAbsorb = max(int(ZOOM_FACTOR_ULTRAMAP*player_radius(main_player)**2),1)
    
        for u in range( int(Xcenter-4), int(Xcenter+4)+1 ):
            for v in range( int(Ycenter-4), int(Ycenter+4)+1 ):
                if ((u>=0 and u<65) and (v>=0 and v<31)):
                    if not cango(main_player,u,v): #((getBlockBottom(u,v)<=main_player[P_Z] or getBlockBottom(u,v)<=player_height(main_player)) and getBlockHeight(u,v)>main_player[P_Z] ):
                        self.ultramap.create_rectangle(X_CENTER_ULTRA_MAP+(u-Xcenter)*ZOOM_FACTOR_ULTRAMAP,
                                                      Y_CENTER_ULTRA_MAP+(v-Ycenter)*ZOOM_FACTOR_ULTRAMAP,
                                                      X_CENTER_ULTRA_MAP+(u-Xcenter+1)*ZOOM_FACTOR_ULTRAMAP,
                                                      Y_CENTER_ULTRA_MAP+(v-Ycenter+1)*ZOOM_FACTOR_ULTRAMAP, fill="grey", outline="black")

        for player in self.pyka3d.get_players():
            if player==main_player:
                self.ultramap.create_oval( X_CENTER_ULTRA_MAP-RadiusPlayer, Y_CENTER_ULTRA_MAP-RadiusPlayer,
                                         X_CENTER_ULTRA_MAP+RadiusPlayer, Y_CENTER_ULTRA_MAP+RadiusPlayer, outline="#0000FF", width=2 )        
                
                self.ultramap.create_oval( X_CENTER_ULTRA_MAP-RadiusPlayerAbsorb, Y_CENTER_ULTRA_MAP-RadiusPlayerAbsorb,
                                         X_CENTER_ULTRA_MAP+RadiusPlayerAbsorb, Y_CENTER_ULTRA_MAP+RadiusPlayerAbsorb, fill="#0000FF" )          

                self.ultramap.create_line( X_CENTER_ULTRA_MAP, Y_CENTER_ULTRA_MAP,
                                         X_CENTER_ULTRA_MAP+VECTOR_ULTRA*player_step(main_player)*cos(main_player[P_ANGLE]), Y_CENTER_ULTRA_MAP+VECTOR_ULTRA*player_step(main_player)*sin(main_player[P_ANGLE]), fill="#0000FF", width=2 ) 
                 
                self.ultramap.create_text(X_CENTER_ULTRA_MAP+20, Y_CENTER_ULTRA_MAP+20, fill="#0000FF", text=str(main_player[P_LEVEL]))
                
                
            elif player!=main_player and (player[P_Z]==main_player[P_Z] or (player[P_Z]<=main_player[P_Z]+2 and player[P_VZ]>0)):
                deltaX0 = player[P_X]-Xcenter
                deltaY0 = player[P_Y]-Ycenter

                RadiusMice = int(ZOOM_FACTOR_ULTRAMAP*player_radius(player))+1
                if abs(deltaX0)<=8 and abs(deltaY0)<=8:
                    if main_player[P_LEVEL] == player[P_LEVEL]:
                        colorscheme = colorMouseLvl
                    elif main_player[P_LEVEL] == player[P_LEVEL] + 1:
                        colorscheme = colorMouselvlminus1
                    elif main_player[P_LEVEL] == player[P_LEVEL] + 2:
                        colorscheme = colorMouselvlminus2      
                    elif main_player[P_LEVEL] == player[P_LEVEL] + 3:
                        colorscheme = colorMouselvlminus3        
                    elif main_player[P_LEVEL] == player[P_LEVEL] + 4:
                        colorscheme = colorMouselvlminus4
                    elif main_player[P_LEVEL] > player[P_LEVEL] + 4:
                        colorscheme = colorMouseLow
                    else:
                        colorscheme = colorMouseHigh  
                        
                    self.ultramap.create_oval( X_CENTER_ULTRA_MAP+deltaX0*ZOOM_FACTOR_ULTRAMAP-RadiusMice,
                                             Y_CENTER_ULTRA_MAP+deltaY0*ZOOM_FACTOR_ULTRAMAP-RadiusMice,
                                                 X_CENTER_ULTRA_MAP+deltaX0*ZOOM_FACTOR_ULTRAMAP+RadiusMice,
                                                 Y_CENTER_ULTRA_MAP+deltaY0*ZOOM_FACTOR_ULTRAMAP+RadiusMice, outline=colorscheme, width=2 )     
                    
                    #self.ultramap.create_line( X_CENTER_ULTRA_MAP+deltaX0*ZOOM_FACTOR_ULTRAMAP,
                    #                         Y_CENTER_ULTRA_MAP+deltaY0*ZOOM_FACTOR_ULTRAMAP,
                    #                             X_CENTER_ULTRA_MAP+deltaX0*ZOOM_FACTOR_ULTRAMAP+VECTOR_ULTRA*player_step(player)*cos(player[P_ANGLE]),
                    #                             Y_CENTER_ULTRA_MAP+deltaY0*ZOOM_FACTOR_ULTRAMAP+VECTOR_ULTRA*player_step(player)*sin(player[P_ANGLE]), fill=colorscheme, width=2 ) 
 
                    self.ultramap.create_line( X_CENTER_ULTRA_MAP+deltaX0*ZOOM_FACTOR_ULTRAMAP,
                                             Y_CENTER_ULTRA_MAP+deltaY0*ZOOM_FACTOR_ULTRAMAP,
                                                 X_CENTER_ULTRA_MAP+deltaX0*ZOOM_FACTOR_ULTRAMAP+VECTOR_ULTRA*(player[NEXT_X]-player[PREV_X]),
                                                 Y_CENTER_ULTRA_MAP+deltaY0*ZOOM_FACTOR_ULTRAMAP+VECTOR_ULTRA*(player[NEXT_Y]-player[PREV_Y]), fill=colorscheme, width=2 ) 
                     
                    self.ultramap.create_text(X_CENTER_ULTRA_MAP+deltaX0*ZOOM_FACTOR_ULTRAMAP+20,
                                             Y_CENTER_ULTRA_MAP+deltaY0*ZOOM_FACTOR_ULTRAMAP+20,
                                             fill=colorscheme, text=str(player[P_LEVEL]))


    def update_zoommap(self):       
        self.init_zoommap()

    def update_ultramap(self):       
        self.init_ultramap()

    def update_minimap(self):       
        global NbFusion
        self.init_minimap()
        for i in self.minimap_pykas:
            self.minimap.delete(i)
        self.minimap_pykas.clear()
        for i in self.pyka3d.get_players():
            #r = player_radius(i) ** 2*EXTRA_ZOOM_ENTITIES
            if i == self.pyka3d.get_main_player():
                self.minimap_pykas.append(self.minimap.create_rectangle((i[P_X] - 0.25) * ZOOM_FACTOR_MINIMAP, (i[P_Y] - 0.25) * ZOOM_FACTOR_MINIMAP, (i[P_X] + 0.25) * ZOOM_FACTOR_MINIMAP, (i[P_Y] + 0.25) * ZOOM_FACTOR_MINIMAP, fill="#0000FF", outline=""))
                self.minimap_pykas.append(self.minimap.create_line(i[P_X] * ZOOM_FACTOR_MINIMAP, i[P_Y] * ZOOM_FACTOR_MINIMAP, (i[P_X]+2*cos(i[P_ANGLE])*player_step(self.pyka3d.get_main_player())*EXTRA_ZOOM_ENTITIES) * ZOOM_FACTOR_MINIMAP, (i[P_Y]+2*sin(i[P_ANGLE])*player_step(self.pyka3d.get_main_player())*EXTRA_ZOOM_ENTITIES) * ZOOM_FACTOR_MINIMAP, fill="#0000FF", width=2 ))
                self.minimap_pykas.append(self.minimap.create_text((i[P_X]+0.35) * ZOOM_FACTOR_MINIMAP, (i[P_Y]+0.35) * ZOOM_FACTOR_MINIMAP, fill="#0000FF", text=str(i[P_LEVEL]) ))

            else:
                #if i[P_Z] < 1:
                    self.draw_pyka(i)   
            
            main_player = self.pyka3d.get_main_player()
            #self.minimap_pykas.append(self.minimap.create_text(300, MAP_HEIGHT * ZOOM_FACTOR_MINIMAP+20, fill="#FF0000", 
            #                                                   text=str(round(main_player[P_SCORE_IA],1)) + "   Lv=" + str(main_player[P_LEVEL]) +
            #                                                   "     X=" + str(round(main_player[P_X],3)) + "     Y=" + str(round(main_player[P_Y],3))+
            #                                                   "     Z=" + str(round(main_player[P_Z],3))+ "     St=" + str(round(player_step(main_player),3))+
            #                                                   "     Jp=" + str(round(player_jump(main_player),3))+ "     Rd=" + str(round(player_radius(main_player),3)) ))

        self.output.configure(background='white', foreground='black')
        self.output.config(font=('Helvetica',9))
        self.output.delete('1.0', tk.END)
        self.output.insert(tk.END, "INFO PLAYER : Score=" + str(round(main_player[P_SCORE_IA],4)) + "   Lv=" + str(main_player[P_LEVEL]) +
                           "     X=" + str(round(main_player[P_X],3)) + "     Y=" + str(round(main_player[P_Y],3))+
                           "     Z=" + str(round(main_player[P_Z],3))+ "     St=" + str(round(player_step(main_player),3))+
                           "     Jp=" + str(round(player_jump(main_player),3))+ "     Rd=" + str(round(player_radius(main_player),3)) )
    
        self.output.insert(tk.END, '\n' + "REMAINING MICE : " + str(len(self.pyka3d.get_players())-1) )
        
        
        
        if NbFusion!=len(self.pyka3d.get_Fusion()):
            Fusions = self.pyka3d.get_Fusion()[NbFusion:]
            print(Fusions)
            for k in range(len(Fusions)):
                self.output.insert(tk.END, '\n' + "FUSION OF MICE : " + str(Fusions[k][0]) + " AND " + str(Fusions[k][1]) )
                
                self.minimap_pykas.append(self.minimap.create_oval((Fusions[k][2] - 1) * ZOOM_FACTOR_MINIMAP,
                                                               (Fusions[k][3] - 1) * ZOOM_FACTOR_MINIMAP,
                                                               (Fusions[k][2] + 1) * ZOOM_FACTOR_MINIMAP,
                                                               (Fusions[k][3] + 1) * ZOOM_FACTOR_MINIMAP,
                                                               outline="#FFFFFF", width=4))
            NbFusion = len(self.pyka3d.get_Fusion())
            self.output.configure(background='red', foreground='white')

            
        players = self.pyka3d.get_players()
        players = players[1:]
        Niveaux = sorted([elt[P_LEVEL] for elt in players])
        self.output.insert(tk.END, '\n' + "LEVELS OF MICE = [ " )
        for i in range(len(Niveaux)):
            self.output.insert(tk.END, str(Niveaux[i]))
            if i!=len(Niveaux)-1:
                self.output.insert(tk.END, "," )    
        self.output.insert(tk.END, " ]" )
        
        self.output.insert(tk.END, 
                           '\n'+"Nb de souris de niv > niv pivot : " + str(sum(Niveaux.count(k) for k in range(4+1,25))-1) +
                           "   max : "+str(16-(self.pyka3d.get_main_player()[P_LEVEL]-3)) +'\n'                          
                           )
        self.outputfusions.delete("1.0", 'end')
        self.outputfusions.config(font=('Helvetica',8))
        Fusions = self.pyka3d.get_Fusion()
        for k in range(len(Fusions)):
            self.outputfusions.insert(tk.END, "FUSION OF MICE : " + str(Fusions[k][0]) + " AND " + str(Fusions[k][1]) + '\n')
        

    def ctrl_move_length_changed(self, *args):
        global move_length, reload
        try:
            move_length =  float(self.move_length.get())
            reload = True
        except ValueError:
            pass

    def ctrl_new_orientation_changed(self, *args):
        global angle, reload
        try:
            angle =  float(self.angle.get()) * pi / 180
            reload = True
        except ValueError:
            pass

    def ctrl_dump_list(self):
        optimized = []

        for i in self.actions:
            if len(optimized) == 0:
                optimized.append(list(i))
                continue
            
            if i[0] in [0, 1, 2, 3, 4]:
                if i[0] == optimized[-1][0]:
                    optimized[-1][1] += i[1]
                    continue

            optimized.append(list(i))
        print(optimized)
        
        now = datetime.now()
        filename = "Dump_Pyka_"+str(round(self.pyka3d.get_players()[0][6],4))+"_"+now.strftime("%d-%m-%Y_%H-%M-%S")+".txt"
        f = open(filename, "w")
        f.write("l=[")
            
        for i in range(len(optimized)):
            if i%15==0: f.write('\n')
            f.write("["+str(optimized[i][0])+","+str(optimized[i][1])+"]")
            if i!=len(optimized)-1: f.write(",")
        
        f.write('\n')
        f.write("]")
        
        f.close()
        

    def ctrl_reset(self):
        global line
        self.pyka3d.init()
        self.output.configure(background='white', foreground='black')
        self.ctrl_reload()
        self.ctrl_refresh()
        self.update_minimap()
        self.update_zoommap()
        self.update_ultramap()
        self.actions = []
        self.history = []
        self.history.append(self.pyka3d.save())
        line = [(self.pyka3d.get_players()[0][0], self.pyka3d.get_players()[0][2] / 10, self.pyka3d.get_players()[0][1])]

    def ctrl_load_game(self):
        try:
            l = json.loads(self.load_game.get())
            self.ctrl_reset()
            self.disable_updates = True
            while len(l) != 0:
                if (len(l) <= 2):
                    self.disable_updates = False
                a = l.pop(0)
                if isinstance(a, list) or isinstance(a, tuple):
                    self.ctrl_action(a[0], a[1])
                else:
                    self.ctrl_action(a, l.pop(0))
        except:
            print("ERROR LOADING")
            self.disable_updates = False

    def ctrl_reload(self):
        global reload, line
        line.append((self.pyka3d.get_players()[0][0], self.pyka3d.get_players()[0][2] / 10, self.pyka3d.get_players()[0][1]))
        reload = True
    
    def ctrl_refresh(self):
        self.level.set("Level: " + str(self.pyka3d.get_players()[0][5]))
        self.score.set("Score: " + str(self.pyka3d.get_players()[0][6]))
        self.souris.set("Souris: " + str(len(self.pyka3d.get_players())-1))
        self.update_minimap()
        self.update_zoommap()
        self.update_ultramap()
        self.dummy.focus_set()

    def ctrl_undo(self):
        global line, reload
        
        if (len(self.history) == 1):
            return
        self.history.pop()
        if (len(self.actions) != 0):
            self.actions.pop()
        self.pyka3d.load(self.history[-1])
        line.pop()
        line.pop()
        
        NbMice=len(self.pyka3d.get_players())-1
        NbFuse=len(self.pyka3d.get_Fusion())
        print("UNDO : NbMice=",NbMice," NbFusion=",NbFuse," Remove= ",NbMice+NbFuse-99, "entries")
        if (NbMice+NbFuse-99)>0:
            print( "Need to remove ",NbMice+NbFuse-99, "fusions in the list" )
            self.pyka3d.clear_Fusion(NbMice+NbFuse-99)
        
        self.ctrl_reload()
        self.ctrl_refresh()
        self.update_minimap()
        self.update_zoommap()
        self.update_ultramap()
        

    def ctrl_action(self, id, param):
        self.actions.append((id, param))
        self.pyka3d.action([id, param])
        self.history.append(self.pyka3d.save())
        self.ctrl_reload()
        self.ctrl_refresh()

    def ctrl_keyboard(self, event):
        print("Keypressed : ", event.keysym )
        if event.keysym in self.keybinds:
            self.keybinds[event.keysym]()

    def ctrl_jump(self):
        self.ctrl_action(7, 1)

    def ctrl_forward(self):
        self.ctrl_action(0, 1)

    def ctrl_backward(self):
        self.ctrl_action(2, 1)

    def ctrl_left(self):
        self.ctrl_action(5, 3.14159226/8)

    def ctrl_right(self):
        self.ctrl_action(6, 3.14159226/8)

    def ctrl_move_left(self):
        self.ctrl_action(3, 1)

    def ctrl_move_right(self):
        self.ctrl_action(1, 1)

    def ctrl_nop(self):
        self.ctrl_action(4, 1)
    
    def ctrl_custom_rotate(self):
        self.ctrl_action(6, angle - self.pyka3d.get_players()[0][3])
    
    def ctrl_custom_move(self):
        self.ctrl_action(0, move_length)
    
    def ctrl_custom_nop(self):
        try:
            nop_count =  float(self.nop_count.get())
            self.ctrl_action(4, nop_count)
        except ValueError:
            pass


def ctrl_line():
    global line
    return line

def ctrl_should_reload():
    global reload
    return reload

def ctrl_clear_reload():
    global reload
    reload = False

def ctrl_run(instance):
    app = App(instance)
    app.mainloop()

def ctrl_start(instance):
    x = threading.Thread(target=ctrl_run, kwargs={'instance': instance})
    x.start()
