
from math import pi, cos, sin, tan, sqrt, ceil, atan2, floor, asin, acos


import numpy as np
import pygfx as gfx

from pygfx.linalg import Vector3
from wgpu.gui.glfw import WgpuCanvas, run

from ctrl import *

import ctrl
from core.pyka3dlbcli import (Pyka3D, getBlockTop, player_height,
                              player_radius, player_step, P_LEVEL, P_X, P_Y, P_Z, P_ANGLE)
from map import *

instance = Pyka3D()

instance.init()

renderer = gfx.renderers.WgpuRenderer(WgpuCanvas())
scene = gfx.Scene()
gl_players = []
gl_cp = {}

init_map(scene)

player_positions = []


for id, player in enumerate(instance.get_players()):
    p = gfx.Mesh(gfx.cylinder_geometry(),gfx.MeshPhongMaterial(color="#00ff00" if id==0 else "#ffff00", side="front"), )
    
    p.position.set( player[0], player[2] / 10 + (player_height(player) / 20), player[1])
    p.scale.set(player_radius(player)**2, player_radius(player)**2, player_height(player) / 10)
    
    rot = gfx.linalg.Quaternion().set_from_euler(gfx.linalg.Euler(3.141592 / 2, 0, 0))
    
    p.rotation.multiply(rot)
    gl_players.append(p)
    scene.add(p)
    if id == 0:
        player_positions.append( (player[0], player[2] / 10, player[1]) )

for pos in instance.get_cp():
    cp = gfx.Mesh( gfx.box_geometry(0.5, 0.5, 0.5), gfx.MeshPhongMaterial(color="#ff00ff", side="front"),    )
    cp.position.set( pos[0] + 0.5, getBlockTop(pos[0], pos[1]) / 10, pos[1] + 0.5 )
    scene.add(cp)
    gl_cp[pos] = cp

camera = gfx.PerspectiveCamera(50, 16/9)

targetplayer = gl_players[0].position.clone()

controller = gfx.OrbitController(Vector3(camera.position.x, camera.position.y + 0.5, camera.position.z + 0.1), Vector3(camera.position.x, camera.position.y, camera.position.z))
controller.add_default_event_handlers(renderer, camera)

player_orientation = gfx.Line( gfx.Geometry(positions=[(0, 0, 0), (2, 0, 0)]),   gfx.LineMaterial(thickness=2.0, color="#00ffff"), )
scene.add(player_orientation)

new_orientation = gfx.Line( gfx.Geometry(positions=[(0, 0, 0), (2, 0, 0)]),    gfx.LineMaterial(thickness=2.0, color="#ff00ff"), )
scene.add(new_orientation)

player_path = gfx.Line(    gfx.Geometry(positions=[(0, 0, 0), (2, 0, 0)]),    gfx.LineMaterial(thickness=2.0, color="#ff0000"), )
scene.add(player_path)


def reload_players():
    global gl_players#, controller

    for pos in gl_cp.keys():
        gl_cp[pos].material.color = "#ffffff" if pos in instance.get_cp( ) else "#222222"

    for id, player in enumerate(gl_players):
        if (id < len(instance.get_players())):
            
            p = instance.get_players()[id]
            
            player.position.set(p[0], p[2] / 10 + (player_height(p) / 20), p[1])
            player.scale.set(player_radius(p), player_radius(p), player_height(p) / 10)
            
            if id == 0:
                player.material.color = colorPlayer
                player_positions.append( (p[0], p[2] / 10, p[1]) )
                player_path.geometry = gfx.Geometry(positions=ctrl.ctrl_line())

                player_orientation.position.set( p[0], p[2] / 10, p[1] )
                player_orientation.rotation.set_from_euler( gfx.linalg.Euler(0, - p[3], 0) )
                
                new_orientation.position.set(p[0], p[2] / 10, p[1])
                new_orientation.rotation.set_from_euler( gfx.linalg.Euler(0, - ctrl.angle, 0) )
                new_orientation.geometry = gfx.Geometry(positions=[ (0, 0, 0), (ctrl.move_length * player_step(instance.get_players()[0] ), 0, 0) ])
            
            elif instance.get_players()[id][5] == instance.get_players()[0][5]:
                player.material.color = colorMouseLvl
            elif instance.get_players()[id][5] == instance.get_players()[0][5] - 1:
                player.material.color = colorMouselvlminus1
            elif instance.get_players()[id][5] == instance.get_players()[0][5] - 2:
                player.material.color = colorMouselvlminus2
            elif instance.get_players()[id][5] == instance.get_players()[0][5] - 3:
                player.material.color = colorMouselvlminus3
            elif instance.get_players()[id][5] == instance.get_players()[0][5] - 4:
                player.material.color = colorMouselvlminus4
            elif instance.get_players()[id][5] < instance.get_players()[0][5] - 4:
                player.material.color = colorMouseLow
            else:
                player.material.color = colorMouseHigh

        else:
            player.position.set(-1, -1, -1)
            
    controller.target = gl_players[0].position.clone()   
    controller._update_up_quats()
    
reload_players()

def animate(): 
    controller.update_camera(camera)  
    renderer.render(scene, camera)
    renderer.request_draw()

    if (ctrl.ctrl_should_reload()):
        reload_players()
        ctrl.ctrl_clear_reload()


if __name__ == "__main__":
    renderer.request_draw(animate)
    ctrl.ctrl_start(instance)
    run()
